"""
Julia package for easy access to the Ken French Data Library files. See `readFamaFrench`
"""
module FamaFrenchData

using Downloads
using CSV
using ZipFile
using DataFrames
import Dates: now



export readFamaFrench, downloadFamaFrench, listFamaFrench

const KFDLftp = "http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/"

"""
    readFamaFrench(ffn;kwargs...)

`ffn` can be the table name (in which case it is retreived from the web) or a path to the local file.
`kwargs` are passed to `CSV.File`. Missing values (`-99.99` or `-999`) are replaced with `missing`.

Returns three pieces:

    - `tables::Vector{DataFrame}` - the extracted tables

    - `tablenotes::Vector{String}` - any notes to the 
    
    - `filenotes::String` - notes at the top of the file
"""
function readFamaFrench(ffn;kwargs...)
    if !isfile(ffn)
        io = IOBuffer()
        Downloads.download(pathtoFamaFrench(ffn),io)
        z = ZipFile.Reader(io)
        ff = first(z.files)
    else
        ff = open(ffn)
    end
    tables, tablenotes, filenotes = parsefile(ff;kwargs...)
    close(ff)
    return tables, tablenotes, filenotes
end




"""
    downloadFamaFrench(savename,ffn)

Saves the extracted CSV file from `ffn` to local file `savename`.
"""
function downloadFamaFrench(savename,ffn)
    open(savename,"w") do f
        io = IOBuffer()
        Downloads.download(pathtoFamaFrench(ffn),io)
        z = ZipFile.Reader(io)
        ff = first(z.files)
        for s in readlines(ff)
            println(f,s)
        end
        close(ff)
    end
end




"""
    listFamaFrench(;refresh=false)

Returns a vector of possible table names. Reads from `listFamaFrench.txt`.
When `refresh = true`, first crawls the website to find current list of tables, then overwrites `listFamaFrench.txt` with this list.
The selection of tables is rarely changed, so the provided list is likely sufficient.
"""
function listFamaFrench(;refresh::Bool = false)
    LFF_file = abspath(joinpath(@__DIR__, "..", "listFamaFrench.txt"))
    if refresh
        # grabs filenames from website
        wname  = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/data_library.html" 
        io = IOBuffer()
        Downloads.download(wname,io)
        hs = String(take!(io))
        close(io)
        regx = r"(?<=ftp\/)(.*)(?=_CSV)"
        res=collect(eachmatch(regx,hs)) 
        nms = getfield.(res,:match)
        # writes the available filenames to a file
        rm(LFF_file)
        open(LFF_file,"w") do f
            println(f,"# Generated by FamaFrenchData.jl $(now())")
            for s in nms
                println(f,s)
            end
        end
    else
        nms = readlines(LFF_file)
        popfirst!(nms)
    end

    return nms
end




########################################
# UNEXPORTED FUNCTIONS
########################################
"""
    parsefile(lines;kwargs...)

The workhorse function behind `readFamaFrench`.

`lines` is an `IO`.
`kwargs` are passed to `CSV.File`. Missing values (`-99.99` or `-999`) are replaced with `missing`.

Returns three pieces:

    - `tables::Vector{DataFrame}` - the extracted tables

    - `tablenotes::Vector{String}` - any notes to the 
    
    - `filenotes::String` - notes at the top of the file
"""
function parsefile(lines;kwargs...)
    csvopt = (missingstrings = ["-99.99","-999"],normalizenames = false,kwargs...)

    stringarray = readlines(lines,keep=true)
    striparray = strip.(stringarray)
    splits = split.(stringarray,",")
    ncols = length.(splits)

    tranges = contiguousblocks(ncols)
    ntables = length(tranges)

    allempty = findall(isempty,striparray)
    lastempty = [lastornothing(filter(x -> x < k,allempty)) for k in first.(tranges)]
    nameranges = [isnothing(lastempty[i]) ? nothing : range(lastempty[i],first(tranges[i])-1,step=1) for i in 1:ntables] 

    dfvec = Vector{DataFrame}(undef,ntables)
    for i in 1:ntables
        ios = IOBuffer(string(stringarray[tranges[i]]...))
        dfvec[i] = CSV.read(ios,DataFrame;csvopt...)
        rename!(dfvec[i],Symbol(first(names(dfvec[i]))) => :Date)
        close(ios)
    end
    if isnothing(first(nameranges))
        notestop = tranges[1][1]-1
        tablenotes = [i==1 ? "" : strip_rn(string(striparray[nameranges[i]]...)) for i in 1:ntables]
    else
        tablenotes = [strip_rn(string(striparray[nameranges[i]]...)) for i in 1:ntables]
        notestop = first(lastempty)
    end
    filenotes = strip_rn(string(lines.name,"  ",strip.(stringarray[1:notestop])...))

    return dfvec,tablenotes,filenotes
end


strip_rn(s) = foldl(replace,["\r"=>"", "\n"=>" "],init=s)



"""
    contiguousblocks(x)

Takes in a vector of `Int` and returns the portions containing contiguous blocks (sequences that increase by 1).

Requires a sequence to have length 3 or greater.
"""
function contiguousblocks(x::AbstractVector{T}) where {T}
    M = maximum(x)
    L = length(x)
    tranges = Vector{UnitRange{Int64}}(undef,0)
    for i in 2:M
        tmpi = findall(==(i),x)
        isempty(tmpi) && continue
        starti = tmpi[1]
        lasti = tmpi[1]
        for (j,n) in enumerate(tmpi)
            j==1 && continue
            distance = n-lasti
            if distance==1
                lasti = n
                if j==length(tmpi)
                    R = starti:lasti
                    # println(R)
                    length(R)>=3 && push!(tranges,R)
                end
                continue
            else
                R = starti:lasti
                # println(R)
                length(R)>=3 && push!(tranges,R)
                starti = n
                lasti = n
            end
        end

    end
    return tranges
end



"""
    pathtoFamaFrench(ffn)

Generates the full web path to the file.
"""
pathtoFamaFrench(ffn) = joinpath(KFDLftp,string(ffn, "_CSV.zip"))




"""
    lastornothing(x) = isempty(x) ? nothing : last(x)

A helper function.
"""
lastornothing(x) = isempty(x) ? nothing : last(x)



end # module
